import 'package:example/main.dart';
import 'package:flutter/material.dart';
import 'package:flutterx_live_data/flutterx_live_data.dart';
import 'package:flutterx_preferences/flutterx_preferences.dart';

class MutableLiveDataTest extends StatefulWidget {
  const MutableLiveDataTest({super.key});

  @override
  State<MutableLiveDataTest> createState() => _MutableLiveDataTestState();
}

class _MutableLiveDataTestState extends State<MutableLiveDataTest> {
  final MutableLiveData<int> _x = MutableLiveData(value: 7);
  final MutableLiveData<int> _z = MutableLiveData.late();
  late final MediatorLiveData<int> _y = MediatorLiveData()..addTransformSource<int>(_x, (x) => x * 10);
  late final MediatorLiveData<int> _y2 = MediatorLiveData()..addTransformSource2<int, int>(_x, _z, (x, z) => x * z);
  late final MediatorMutableLiveData<int> _ym = MediatorMutableLiveData(value: 0)
    ..addTransformSource<int>(_x, (x) => x * 10);
  final HistoryMutableLiveData<int> _h = HistoryMutableLiveData(value: 0);
  final StorageMutableLiveData<int> _s = StorageMutableLiveData.int(key: 's', defaultValue: 0, storage: sessionStorage);
  final MutableLiveData<bool> _switch = MutableLiveData(value: true);

  @override
  Widget build(BuildContext context) => Row(children: [
        xColumn('Data', [
          const Text('x [mutable]'),
          const Text('z [mutable,late]'),
          const Text('y [mediator =x*10]'),
          const Text('y2 [mutable,mediator(1) =x*z]'),
          const Text('ym [mutable,mediator(2) =x*10]'),
          const Text('h [mutable,history]'),
          const Text('s [mutable,storage]'),
          LiveDataBuilder<bool>(
              data: _switch, builder: (context, s) => Text('widget [combined(2) =x+${s ? 'y' : 'z'}]')),
        ]),
        const SizedBox(width: kMinInteractiveDimension),
        xColumn('Value', [
          xData(_x),
          xData(_z),
          xData(_y),
          xData(_y2),
          xData(_ym),
          xData(_h),
          xData(_s),
          LiveDataBuilder<bool>(
              data: _switch,
              builder: (context, s) => CombinedLiveDataBuilder.with2<int, int, int>(
                  x1: _x,
                  x2: s ? _y : _z,
                  transform: (x, y) => x + y,
                  builder: (context, result) => Text('$result'),
                  placeholder: const Text('--'))),
        ]),
        const SizedBox(width: kMinInteractiveDimension),
        xColumn('Actions', [
          xAction(onPressed: () => _x.value++, icon: Icons.add),
          xAction(onPressed: () => _z.initialized ? _z.value++ : _z.value = 0, icon: Icons.add),
          const SizedBox(height: 20),
          const SizedBox(height: 20),
          xAction(onPressed: () => _ym.initialized ? _ym.value++ : _ym.value = 0, icon: Icons.add),
          xActions([
            xAction(onPressed: () => _h.value++, icon: Icons.add),
            LiveDataBuilder<bool>(
                data: _h.canUndo,
                builder: (context, enabled) => xAction(onPressed: enabled ? _h.historyUndo : null, icon: Icons.undo)),
            LiveDataBuilder<bool>(
                data: _h.canRedo,
                builder: (context, enabled) => xAction(onPressed: enabled ? _h.historyRedo : null, icon: Icons.redo)),
          ]),
          xActions([
            xAction(onPressed: () => _s.value++, icon: Icons.add),
            xAction(onPressed: _s.reset, icon: Icons.delete),
          ]),
          xAction(onPressed: () => _switch.value = !_switch.value, icon: Icons.swap_horiz),
        ]),
      ]);
}
