// ignore_for_file: collection_methods_unrelated_type

part of '../flutterx_live_data.dart';

abstract class LiveCollection<C> extends ObservableData<C> {
  LiveCollection(C source, {required super.equality}) {
    super._setData(source);
  }

  void setTo(C value);

  void postSetTo(FutureOr<C> value) => _postSetData(value);

  bool apply(bool Function(C data) update) {
    if (update(_data)) {
      invalidate();
      return true;
    }
    return false;
  }

  @override
  void _setData(C value) => setTo(value);

  @override
  String toString() => _data.toString();
}

class LiveList<E> extends LiveCollection<List<E>> implements List<E> {
  final Equality<E> elementEquality;

  LiveList({List<E>? source, Equality<E>? elementEquality})
      : elementEquality = elementEquality ??= IdentityEquality<E>(),
        super(source ?? <E>[], equality: ListEquality(elementEquality));

  @override
  void setTo(List<E> value) {
    if (equality.equals(_data, value)) return;
    _data
      ..clear()
      ..addAll(value);
    invalidate();
  }

  @override
  bool any(bool Function(E) test) => _data.any(test);

  @override
  List<T> cast<T>() => _data.cast<T>();

  @override
  bool contains(Object? element) => _data.contains(element);

  @override
  E elementAt(int index) => _data.elementAt(index);

  @override
  bool every(bool Function(E) test) => _data.every(test);

  @override
  Iterable<T> expand<T>(Iterable<T> Function(E) f) => _data.expand(f);

  @override
  E get first => _data.first;

  @override
  E firstWhere(bool Function(E) test, {E Function()? orElse}) => _data.firstWhere(test, orElse: orElse);

  @override
  T fold<T>(T initialValue, T Function(T previousValue, E element) combine) => _data.fold(initialValue, combine);

  @override
  Iterable<E> followedBy(Iterable<E> other) => _data.followedBy(other);

  @override
  void forEach(void Function(E) f) => _data.forEach(f);

  @override
  bool get isEmpty => _data.isEmpty;

  @override
  bool get isNotEmpty => _data.isNotEmpty;

  @override
  Iterator<E> get iterator => _data.iterator;

  @override
  String join([String separator = '']) => _data.join(separator);

  @override
  E get last => _data.last;

  @override
  E lastWhere(bool Function(E) test, {E Function()? orElse}) => _data.lastWhere(test, orElse: orElse);

  @override
  int get length => _data.length;

  @override
  Iterable<T> map<T>(T Function(E) f) => _data.map(f);

  @override
  E reduce(E Function(E value, E element) combine) => _data.reduce(combine);

  @override
  E get single => _data.single;

  @override
  E singleWhere(bool Function(E) test, {E Function()? orElse}) => _data.singleWhere(test, orElse: orElse);

  @override
  Iterable<E> skip(int n) => _data.skip(n);

  @override
  Iterable<E> skipWhile(bool Function(E) test) => _data.skipWhile(test);

  @override
  Iterable<E> take(int n) => _data.take(n);

  @override
  Iterable<E> takeWhile(bool Function(E) test) => _data.takeWhile(test);

  @override
  List<E> toList({bool growable = true}) => _data.toList(growable: growable);

  @override
  Set<E> toSet() => _data.toSet();

  @override
  Iterable<E> where(bool Function(E) test) => _data.where(test);

  @override
  Iterable<T> whereType<T>() => _data.whereType<T>();

  @override
  E operator [](int index) => _data[index];

  @override
  void operator []=(int index, E value) {
    if (elementEquality.equals(_data[index], value)) return;
    _data[index] = value;
    invalidate();
  }

  @override
  List<E> operator +(List<E> other) => _data + other;

  @override
  void add(E value) {
    _data.add(value);
    invalidate();
  }

  @override
  void addAll(Iterable<E> iterable) {
    if (iterable.isEmpty) return;
    _data.addAll(iterable);
    invalidate();
  }

  @override
  Map<int, E> asMap() => _data.asMap();

  @override
  void clear() {
    if (_data.isEmpty) return;
    _data.clear();
    invalidate();
  }

  @override
  void fillRange(int start, int end, [E? fillValue]) {
    if (start == end) return;
    _data.fillRange(start, end, fillValue);
    invalidate();
  }

  @override
  set first(E value) {
    if (elementEquality.equals(_data.first, value)) return;
    _data.first = value;
    invalidate();
  }

  @override
  Iterable<E> getRange(int start, int end) => _data.getRange(start, end);

  @override
  int indexOf(E element, [int start = 0]) => _data.indexOf(element, start);

  @override
  int indexWhere(bool Function(E) test, [int start = 0]) => _data.indexWhere(test, start);

  @override
  void insert(int index, E element) {
    _data.insert(index, element);
    invalidate();
  }

  @override
  void insertAll(int index, Iterable<E> iterable) {
    if (iterable.isEmpty) return;
    _data.insertAll(index, iterable);
    invalidate();
  }

  @override
  set last(E value) {
    if (elementEquality.equals(_data.last, value)) return;
    _data.last = value;
    invalidate();
  }

  @override
  int lastIndexOf(E element, [int? start]) => _data.lastIndexOf(element, start);

  @override
  int lastIndexWhere(bool Function(E) test, [int? start]) => _data.lastIndexWhere(test, start);

  @override
  set length(int newLength) {
    if (_data.length == newLength) return;
    _data.length = newLength;
    invalidate();
  }

  @override
  bool remove(Object? value) {
    final result = _data.remove(value);
    if (result) invalidate();
    return result;
  }

  @override
  E removeAt(int index) {
    final result = _data.removeAt(index);
    invalidate();
    return result;
  }

  @override
  E removeLast() {
    final result = _data.removeLast();
    invalidate();
    return result;
  }

  @override
  void removeRange(int start, int end) {
    if (start == end) return;
    _data.removeRange(start, end);
    invalidate();
  }

  @override
  void removeWhere(bool Function(E) test) {
    final length = _data.length;
    _data.removeWhere(test);
    if (length != _data.length) invalidate();
  }

  @override
  void replaceRange(int start, int end, Iterable<E> iterable) {
    if (start == end && iterable.isEmpty) return;
    _data.replaceRange(start, end, iterable);
    invalidate();
  }

  @override
  void retainWhere(bool Function(E) test) {
    final length = _data.length;
    _data.retainWhere(test);
    if (length != _data.length) invalidate();
  }

  @override
  Iterable<E> get reversed => _data.reversed;

  @override
  void setAll(int index, Iterable<E> iterable) {
    if (index == _data.length || iterable.isEmpty) return;
    _data.setAll(index, iterable);
    invalidate();
  }

  @override
  void setRange(int start, int end, Iterable<E> iterable, [int skipCount = 0]) {
    if (start == end) return;
    _data.setRange(start, end, iterable, skipCount);
    invalidate();
  }

  @override
  void shuffle([Random? random]) {
    _data.shuffle(random);
    invalidate();
  }

  @override
  void sort([int Function(E, E)? compare]) {
    _data.sort(compare);
    invalidate();
  }

  @override
  List<E> sublist(int start, [int? end]) => _data.sublist(start, end);
}

class HistoryLiveList<E> extends LiveList<E> with HistoryMixin<List<E>> {
  @override
  final StringEncode<List<E>>? encode;

  @override
  final StringDecode<List<E>>? decode;

  HistoryLiveList({super.source, this.encode, this.decode});
}

class MediatorLiveList<E> extends LiveList<E> with MediatorMixin<List<E>> {
  MediatorLiveList({super.source, super.elementEquality, bool reactive = false}) {
    this.reactive = reactive;
  }
}

class StorageLiveList<E> extends LiveList<E> with StorageMixin<List<E>> {
  @override
  final String key;

  @override
  final List<E> defaultValue;

  @override
  final Map<String, String> storage;

  @override
  final StringEncode<List<E>> encode;

  StorageLiveList({
    required this.key,
    this.defaultValue = const [],
    required this.storage,
    this.encode = jsonEncode,
    StringDecode<Iterable<E>>? decode,
  }) : super(source: List<E>.of(StorageMixin.resolve(key, defaultValue, storage, decode ?? StorageMixin.decodeList)));

  StorageLiveList.enumeration({
    required String key,
    List<E> defaultValue = const [],
    required Map<String, String> storage,
    required Iterable<E> values,
  }) : this(
            key: key,
            defaultValue: defaultValue,
            storage: storage,
            encode: StorageMixin.encodeEnumIterable,
            decode: (source) => StorageMixin.decodeEnumIterable<E>(source, values));

  StorageLiveList.jsonObject({
    required String key,
    List<E> defaultValue = const [],
    required Map<String, String> storage,
    dynamic Function(E value)? toJson,
    required E Function(Map<String, dynamic> json) fromJson,
  }) : this(
            key: key,
            defaultValue: defaultValue,
            storage: storage,
            encode: (value) => StorageMixin.encodeJsonIterable(toJson == null ? value : value.map(toJson)),
            decode: (source) => StorageMixin.decodeJsonIterable<E>(source, (item) => fromJson(item)));
}

class LiveMap<K, V> extends LiveCollection<Map<K, V>> implements Map<K, V> {
  final Equality<V?> valueEquality;

  LiveMap({Map<K, V>? source, Equality<K>? keyEquality, Equality<V>? valueEquality})
      : valueEquality = valueEquality ??= IdentityEquality<V>(),
        super(source ?? <K, V>{}, equality: MapEquality(keys: keyEquality ?? Equality<K>(), values: valueEquality));

  @override
  void setTo(Map<K, V> value) {
    if (value.length == _data.length) {
      if (identical(value, _data)) return;
      var equals = true;
      for (final key in value.keys) {
        if (!_data.containsKey(key) || !valueEquality.equals(_data[key], value[key])) {
          equals = false;
          break;
        }
      }
      if (equals) return;
    }
    _data
      ..clear()
      ..addAll(value);
    invalidate();
  }

  @override
  V? operator [](Object? key) => _data[key];

  @override
  void operator []=(K key, V value) {
    if (_data.containsKey(key) && valueEquality.equals(_data[key], value)) return;
    _data[key] = value;
    invalidate();
  }

  @override
  void addAll(Map<K, V> other) {
    if (other.isEmpty) return;
    if (other.length <= _data.length) {
      if (identical(other, _data)) return;
      var equals = true;
      for (final key in other.keys) {
        if (!_data.containsKey(key) || !valueEquality.equals(_data[key], other[key])) {
          equals = false;
          break;
        }
      }
      if (equals) return;
    }
    _data.addAll(other);
    invalidate();
  }

  @override
  void addEntries(Iterable<MapEntry<K, V>> entries) {
    if (entries.isEmpty) return;
    _data.addEntries(entries);
    invalidate();
  }

  @override
  void clear() {
    if (_data.isEmpty) return;
    _data.clear();
    invalidate();
  }

  @override
  Map<K2, V2> cast<K2, V2>() => _data.cast<K2, V2>();

  @override
  bool containsKey(Object? key) => _data.containsKey(key);

  @override
  bool containsValue(Object? value) => _data.containsValue(value);

  @override
  Iterable<MapEntry<K, V>> get entries => _data.entries;

  @override
  void forEach(void Function(K, V) f) => _data.forEach(f);

  @override
  bool get isEmpty => _data.isEmpty;

  @override
  bool get isNotEmpty => _data.isNotEmpty;

  @override
  Iterable<K> get keys => _data.keys;

  @override
  int get length => _data.length;

  @override
  Map<K2, V2> map<K2, V2>(MapEntry<K2, V2> Function(K, V) transform) => _data.map(transform);

  @override
  V putIfAbsent(K key, V Function() ifAbsent) {
    var updated = false;
    final result = _data.putIfAbsent(key, () {
      updated = true;
      return ifAbsent();
    });
    if (updated) invalidate();
    return result;
  }

  @override
  V? remove(Object? key) {
    final length = _data.length;
    final result = _data.remove(key);
    if (length != _data.length) invalidate();
    return result;
  }

  @override
  void removeWhere(bool Function(K, V) test) {
    final length = _data.length;
    _data.removeWhere(test);
    if (length != _data.length) invalidate();
  }

  @override
  Iterable<V> get values => _data.values;

  @override
  V update(K key, V Function(V) update, {V Function()? ifAbsent}) {
    var updated = false;
    final result = _data.update(key, (value) {
      final newValue = update(value);
      updated = !valueEquality.equals(value, newValue);
      return newValue;
    },
        ifAbsent: ifAbsent == null
            ? null
            : () {
                updated = true;
                return ifAbsent();
              });
    if (updated) invalidate();
    return result;
  }

  @override
  void updateAll(V Function(K, V) update) {
    var updated = false;
    _data.updateAll((key, value) {
      final newValue = update(key, value);
      updated |= !valueEquality.equals(value, newValue);
      return newValue;
    });
    if (updated) invalidate();
  }
}

class HistoryLiveMap<K, V> extends LiveMap<K, V> with HistoryMixin<Map<K, V>> {
  @override
  final StringEncode<Map<K, V>>? encode;

  @override
  final StringDecode<Map<K, V>>? decode;

  HistoryLiveMap({super.source, this.encode, this.decode});
}

class MediatorLiveMap<K, V> extends LiveMap<K, V> with MediatorMixin<Map<K, V>> {
  MediatorLiveMap({super.source, super.keyEquality, super.valueEquality, bool reactive = false}) {
    this.reactive = reactive;
  }
}

class StorageLiveMap<K, V> extends LiveMap<K, V> with StorageMixin<Map<K, V>> {
  @override
  final String key;

  @override
  final Map<K, V> defaultValue;

  @override
  final Map<String, String> storage;

  @override
  final StringEncode<Map<K, V>> encode;

  StorageLiveMap({
    required this.key,
    this.defaultValue = const {},
    required this.storage,
    this.encode = jsonEncode,
    StringDecode<Map<K, V>>? decode,
  }) : super(source: Map<K, V>.of(StorageMixin.resolve(key, defaultValue, storage, decode ?? StorageMixin.decodeMap)));
}

class LiveSet<E> extends LiveCollection<Set<E>> implements Set<E> {
  LiveSet({Set<E>? source, Equality<E>? elementEquality})
      : super(source ?? <E>{}, equality: SetEquality(elementEquality ?? IdentityEquality<E>()));

  @override
  void setTo(Set<E> value) {
    if (value.length == _data.length) {
      if (identical(value, _data)) return;
      var equals = true;
      for (final item in value)
        if (!_data.contains(item)) {
          equals = false;
          break;
        }
      if (equals) return;
    }
    _data
      ..clear()
      ..addAll(value);
    invalidate();
  }

  @override
  bool any(bool Function(E) test) => _data.any(test);

  @override
  Set<T> cast<T>() => _data.cast<T>();

  @override
  bool contains(Object? element) => _data.contains(element);

  @override
  E elementAt(int index) => _data.elementAt(index);

  @override
  bool every(bool Function(E) test) => _data.every(test);

  @override
  Iterable<T> expand<T>(Iterable<T> Function(E) f) => _data.expand(f);

  @override
  E get first => _data.first;

  @override
  E firstWhere(bool Function(E) test, {E Function()? orElse}) => _data.firstWhere(test, orElse: orElse);

  @override
  T fold<T>(T initialValue, T Function(T previousValue, E element) combine) => _data.fold(initialValue, combine);

  @override
  Iterable<E> followedBy(Iterable<E> other) => _data.followedBy(other);

  @override
  void forEach(void Function(E) f) => _data.forEach(f);

  @override
  bool get isEmpty => _data.isEmpty;

  @override
  bool get isNotEmpty => _data.isNotEmpty;

  @override
  Iterator<E> get iterator => _data.iterator;

  @override
  String join([String separator = '']) => _data.join(separator);

  @override
  E get last => _data.last;

  @override
  E lastWhere(bool Function(E) test, {E Function()? orElse}) => _data.lastWhere(test, orElse: orElse);

  @override
  int get length => _data.length;

  @override
  Iterable<T> map<T>(T Function(E) f) => _data.map(f);

  @override
  E reduce(E Function(E value, E element) combine) => _data.reduce(combine);

  @override
  E get single => _data.single;

  @override
  E singleWhere(bool Function(E) test, {E Function()? orElse}) => _data.singleWhere(test, orElse: orElse);

  @override
  Iterable<E> skip(int n) => _data.skip(n);

  @override
  Iterable<E> skipWhile(bool Function(E) test) => _data.skipWhile(test);

  @override
  Iterable<E> take(int n) => _data.take(n);

  @override
  Iterable<E> takeWhile(bool Function(E) test) => _data.takeWhile(test);

  @override
  List<E> toList({bool growable = true}) => _data.toList(growable: growable);

  @override
  Set<E> toSet() => _data.toSet();

  @override
  Iterable<E> where(bool Function(E) test) => _data.where(test);

  @override
  Iterable<T> whereType<T>() => _data.whereType<T>();

  @override
  bool add(E value) {
    final result = _data.add(value);
    if (result) invalidate();
    return result;
  }

  @override
  void addAll(Iterable<E> iterable) {
    if (iterable.isEmpty) return;
    final length = _data.length;
    _data.addAll(iterable);
    if (length != _data.length) invalidate();
  }

  @override
  void clear() {
    if (_data.isEmpty) return;
    _data.clear();
    invalidate();
  }

  @override
  bool containsAll(Iterable<Object?> other) => _data.containsAll(other);

  @override
  Set<E> difference(Set<Object?> other) => _data.difference(other);

  @override
  Set<E> intersection(Set<Object?> other) => _data.intersection(other);

  @override
  E? lookup(Object? element) => _data.lookup(element);

  @override
  bool remove(Object? value) {
    final result = _data.remove(value);
    if (result) invalidate();
    return result;
  }

  @override
  void removeAll(Iterable<Object?> elements) {
    final length = _data.length;
    _data.removeAll(elements);
    if (length != _data.length) invalidate();
  }

  @override
  void removeWhere(bool Function(E) test) {
    final length = _data.length;
    _data.removeWhere(test);
    if (length != _data.length) invalidate();
  }

  @override
  void retainAll(Iterable<Object?> elements) {
    final length = _data.length;
    _data.retainAll(elements);
    if (length != _data.length) invalidate();
  }

  @override
  void retainWhere(bool Function(E) test) {
    final length = _data.length;
    _data.retainWhere(test);
    if (length != _data.length) invalidate();
  }

  @override
  Set<E> union(Set<E> other) => _data.union(other);
}

class HistoryLiveSet<E> extends LiveSet<E> with HistoryMixin<Set<E>> {
  @override
  final StringEncode<Set<E>>? encode;

  @override
  final StringDecode<Set<E>>? decode;

  HistoryLiveSet({super.source, this.encode, this.decode});
}

class MediatorLiveSet<E> extends LiveSet<E> with MediatorMixin<Set<E>> {
  MediatorLiveSet({super.source, super.elementEquality, bool reactive = false}) {
    this.reactive = reactive;
  }
}

class StorageLiveSet<E> extends LiveSet<E> with StorageMixin<Set<E>> {
  @override
  final String key;

  @override
  final Set<E> defaultValue;

  @override
  final Map<String, String> storage;

  @override
  final StringEncode<Set<E>> encode;

  StorageLiveSet({
    required this.key,
    this.defaultValue = const {},
    required this.storage,
    this.encode = StorageMixin.encodeJsonIterable,
    StringDecode<Iterable<E>>? decode,
  }) : super(source: Set<E>.of(StorageMixin.resolve(key, defaultValue, storage, decode ?? StorageMixin.decodeList)));

  StorageLiveSet.enumeration({
    required String key,
    Set<E> defaultValue = const {},
    required Map<String, String> storage,
    required Iterable<E> values,
  }) : this(
            key: key,
            defaultValue: defaultValue,
            storage: storage,
            encode: StorageMixin.encodeEnumIterable,
            decode: (source) => Set.of(StorageMixin.decodeEnumIterable<E>(source, values)));

  StorageLiveSet.jsonObject({
    required String key,
    Set<E> defaultValue = const {},
    required Map<String, String> storage,
    dynamic Function(E value)? toJson,
    required E Function(Map<String, dynamic> json) fromJson,
  }) : this(
            key: key,
            defaultValue: defaultValue,
            storage: storage,
            encode: (value) => StorageMixin.encodeJsonIterable(toJson == null ? value : value.map(toJson)),
            decode: (source) => Set.of(StorageMixin.decodeJsonIterable<E>(source, (item) => fromJson(item))));
}

class LiveQueue<E> extends LiveCollection<Queue<E>> implements Queue<E> {
  LiveQueue({Queue<E>? source, Equality<E>? elementEquality})
      : super(source ?? Queue<E>(), equality: QueueEquality(elementEquality ?? IdentityEquality<E>()));

  @override
  void setTo(Queue<E> value) {
    if (equality.equals(_data, value)) return;
    _data
      ..clear()
      ..addAll(value);
    invalidate();
  }

  @override
  bool any(bool Function(E) test) => _data.any(test);

  @override
  Queue<T> cast<T>() => _data.cast<T>();

  @override
  bool contains(Object? element) => _data.contains(element);

  @override
  E elementAt(int index) => _data.elementAt(index);

  @override
  bool every(bool Function(E) test) => _data.every(test);

  @override
  Iterable<T> expand<T>(Iterable<T> Function(E) f) => _data.expand(f);

  @override
  E get first => _data.first;

  @override
  E firstWhere(bool Function(E) test, {E Function()? orElse}) => _data.firstWhere(test, orElse: orElse);

  @override
  T fold<T>(T initialValue, T Function(T previousValue, E element) combine) => _data.fold(initialValue, combine);

  @override
  Iterable<E> followedBy(Iterable<E> other) => _data.followedBy(other);

  @override
  void forEach(void Function(E) f) => _data.forEach(f);

  @override
  bool get isEmpty => _data.isEmpty;

  @override
  bool get isNotEmpty => _data.isNotEmpty;

  @override
  Iterator<E> get iterator => _data.iterator;

  @override
  String join([String separator = '']) => _data.join(separator);

  @override
  E get last => _data.last;

  @override
  E lastWhere(bool Function(E) test, {E Function()? orElse}) => _data.lastWhere(test, orElse: orElse);

  @override
  int get length => _data.length;

  @override
  Iterable<T> map<T>(T Function(E) f) => _data.map(f);

  @override
  E reduce(E Function(E value, E element) combine) => _data.reduce(combine);

  @override
  E get single => _data.single;

  @override
  E singleWhere(bool Function(E) test, {E Function()? orElse}) => _data.singleWhere(test, orElse: orElse);

  @override
  Iterable<E> skip(int n) => _data.skip(n);

  @override
  Iterable<E> skipWhile(bool Function(E) test) => _data.skipWhile(test);

  @override
  Iterable<E> take(int n) => _data.take(n);

  @override
  Iterable<E> takeWhile(bool Function(E) test) => _data.takeWhile(test);

  @override
  List<E> toList({bool growable = true}) => _data.toList(growable: growable);

  @override
  Set<E> toSet() => _data.toSet();

  @override
  Iterable<E> where(bool Function(E) test) => _data.where(test);

  @override
  Iterable<T> whereType<T>() => _data.whereType<T>();

  @override
  void add(E value) {
    _data.add(value);
    invalidate();
  }

  @override
  void addAll(Iterable<E> iterable) {
    if (iterable.isEmpty) return;
    _data.addAll(iterable);
    invalidate();
  }

  @override
  void addFirst(E value) {
    _data.addFirst(value);
    invalidate();
  }

  @override
  void addLast(E value) {
    _data.addLast(value);
    invalidate();
  }

  @override
  void clear() {
    if (_data.isEmpty) return;
    _data.clear();
    invalidate();
  }

  @override
  bool remove(Object? value) {
    final result = _data.remove(value);
    if (result) invalidate();
    return result;
  }

  @override
  E removeFirst() {
    final result = _data.removeFirst();
    invalidate();
    return result;
  }

  @override
  E removeLast() {
    final result = _data.removeLast();
    invalidate();
    return result;
  }

  @override
  void removeWhere(bool Function(E) test) {
    final length = _data.length;
    _data.removeWhere(test);
    if (length != _data.length) invalidate();
  }

  @override
  void retainWhere(bool Function(E) test) {
    final length = _data.length;
    _data.retainWhere(test);
    if (length != _data.length) invalidate();
  }
}

class HistoryLiveQueue<E> extends LiveQueue<E> with HistoryMixin<Queue<E>> {
  @override
  final StringEncode<Queue<E>>? encode;

  @override
  final StringDecode<Queue<E>>? decode;

  HistoryLiveQueue({super.source, this.encode, this.decode});
}

class MediatorLiveQueue<E> extends LiveQueue<E> with MediatorMixin<Queue<E>> {
  MediatorLiveQueue({super.source, super.elementEquality, bool reactive = false}) {
    this.reactive = reactive;
  }
}

class StorageLiveQueue<E> extends LiveQueue<E> with StorageMixin<Queue<E>> {
  @override
  final String key;

  @override
  final Queue<E> defaultValue;

  @override
  final Map<String, String> storage;

  @override
  final StringEncode<Queue<E>> encode;

  StorageLiveQueue({
    required this.key,
    this.defaultValue = const _EmptyQueue(),
    required this.storage,
    this.encode = StorageMixin.encodeJsonIterable,
    StringDecode<Queue<E>>? decode,
  }) : super(source: Queue.of(StorageMixin.resolve(key, defaultValue, storage, decode ?? StorageMixin.decodeList)));

  StorageLiveQueue.enumeration({
    required String key,
    Queue<E> defaultValue = const _EmptyQueue(),
    required Map<String, String> storage,
    required Iterable<E> values,
  }) : this(
            key: key,
            storage: storage,
            defaultValue: defaultValue,
            encode: StorageMixin.encodeEnumIterable,
            decode: (source) => Queue.of(StorageMixin.decodeEnumIterable<E>(source, values)));

  StorageLiveQueue.jsonObject({
    required String key,
    Queue<E> defaultValue = const _EmptyQueue(),
    required Map<String, String> storage,
    dynamic Function(E value)? toJson,
    required E Function(Map<String, dynamic> json) fromJson,
  }) : this(
            key: key,
            storage: storage,
            defaultValue: defaultValue,
            encode: (value) => StorageMixin.encodeJsonIterable(toJson == null ? value : value.map(toJson)),
            decode: (source) => Queue.of(StorageMixin.decodeJsonIterable<E>(source, (item) => fromJson(item))));
}

class QueueEquality<E> implements Equality<Queue<E>> {
  final IterableEquality<E> _equality;

  QueueEquality([Equality<E> elementEquality = const Equality<Never>()])
      : _equality = IterableEquality(elementEquality);

  @override
  bool equals(Queue<E> queue1, Queue<E> queue2) => _equality.equals(queue1, queue2);

  @override
  int hash(Queue<E> queue) => _equality.hash(queue);

  @override
  bool isValidKey(Object? o) => _equality.isValidKey(o);
}

class _EmptyQueue<E> implements Queue<E> {
  const _EmptyQueue();

  @override
  int get length => 0;

  @override
  Iterator<E> get iterator => Iterable<E>.empty().iterator;

  @override
  dynamic noSuchMethod(Invocation invocation) => super.noSuchMethod(invocation);
}
