part of '../flutterx_live_data.dart';

abstract class Dispatcher {
  static const Dispatcher postFrame = _PostFrameDispatcher();
  static const Dispatcher endOfFrame = _EndOfFrameDispatcher();
  static const Dispatcher microtask = _MicrotaskDispatcher();

  const Dispatcher();

  void post(VoidCallback action);
}

class _PostFrameDispatcher extends Dispatcher {
  const _PostFrameDispatcher();

  @override
  void post(VoidCallback action) => WidgetsBinding.instance.schedulerPhase == SchedulerPhase.persistentCallbacks
      ? WidgetsBinding.instance.addPostFrameCallback((_) => action())
      : action();
}

class _EndOfFrameDispatcher extends Dispatcher {
  const _EndOfFrameDispatcher();

  @override
  void post(VoidCallback action) => WidgetsBinding.instance.endOfFrame.then((_) => action());
}

class _MicrotaskDispatcher extends Dispatcher {
  const _MicrotaskDispatcher();

  @override
  void post(VoidCallback action) => scheduleMicrotask(action);
}
