part of '../flutterx_live_data.dart';

@immutable
class _ImmutableLiveData<T> extends ImmutableLiveData<T> {
  @override
  final T _data;

  @override
  int get version => 0;

  @override
  bool get initialized => true;

  @override
  T? get valueOrNull => _data;

  const _ImmutableLiveData(this._data) : super._();
}

@immutable
class _LateImmutableLiveData<T> extends ImmutableLiveData<T> {
  static late dynamic __data;

  @override
  T get _data => __data;

  @override
  int get version => -1;

  @override
  bool get initialized => false;

  @override
  T? get valueOrNull => null;

  const _LateImmutableLiveData() : super._();
}

@immutable
abstract class ImmutableLiveData<T> implements LiveData<T> {
  static final Map<Object, Set<Object>> _dispatching = {};

  @override
  Equality<T> get equality => Equality<T>();

  @override
  Map<Observer<T>, ObserverWrapper<T>> get _observers => const {};

  @override
  set _data(T value) => throw UnsupportedError('not reachable'); // ignore: avoid_setters_without_getters

  @override
  int get _version => 0;

  @override
  set _version(int value) => throw UnsupportedError('not reachable');

  @override
  int get _activeCount => _dispatching[this]?.length ?? 0;

  @override
  set _activeCount(int value) => throw UnsupportedError('not reachable');

  @override
  bool get active => _activeCount > 0;

  @override
  T get value => _data;

  const factory ImmutableLiveData(T data) = _ImmutableLiveData<T>;

  const factory ImmutableLiveData.late() = _LateImmutableLiveData<T>;

  const ImmutableLiveData._();

  @override
  void onActive() {}

  @override
  void onInactive() {}

  @override
  ObserverWrapper<T> observeForever(Observer<T> onChanged, {Dispatcher? dispatcher}) =>
      addObserverWrapper(ObserverWrapper<T>(onChanged, dispatcher));

  @override
  void removeObserver(Observer<T> observer) => _dispatching.remove(observer);

  @override
  ObserverWrapper<T> addObserverWrapper(ObserverWrapper<T> observer) {
    final dispatcher = observer.dispatcher;
    if (dispatcher == null) {
      observer.onChanged(_data);
    } else {
      final dispatching = (_dispatching[this] ??= {})..add(observer.onChanged);
      dispatcher.post(() => dispatching.remove(observer.onChanged) ? observer.onChanged(_data) : null);
    }
    return observer;
  }

  @override
  void removeObserverWrapper(ObserverWrapper<T> observer) => removeObserver(observer.onChanged);

  @override
  void invalidate() {}

  @override
  void _onObserverAttach(ObserverWrapper<T> observer) => throw UnsupportedError('not reachable');

  @override
  void _onObserverDetach(ObserverWrapper<T> observer) => throw UnsupportedError('not reachable');

  @override
  void _postSetData(FutureOr<T> value) => throw UnsupportedError('not reachable');

  @override
  void _setData(T value) => throw UnsupportedError('not reachable');

  @override
  String toString() => ObservableData.describe(this);
}
