part of '../flutterx_live_data.dart';

class LiveResult<T> extends SelfLiveData<LiveResult<T>> {
  dynamic _value;
  ResultState _state;
  ValueGetter<Future<T>>? _initializer;
  bool _reactive = false;

  ResultState get state => _state;

  T get result => _state == ResultState.success
      ? _value
      : throw StateError('result can only be accessed when state is ResultState.success');

  Object get error => _state == ResultState.error
      ? _value
      : throw StateError('error can only be accessed when state is ResultState.error');

  LiveResult() : _state = ResultState.idle;

  LiveResult.loading() : _state = ResultState.loading;

  LiveResult.error(Object error)
      : _value = error,
        _state = ResultState.error;

  LiveResult.success([T? value])
      : assert(value is T, 'value must be of type $T, got $value'),
        _value = value,
        _state = ResultState.success;

  @override
  void onActive() {
    super.onActive();
    if (_initializer != null && !_reactive && _state == ResultState.idle) reload();
  }

  Future<void> initialize(ValueGetter<Future<T>> initializer, {bool reactive = false}) {
    final changed = _initializer != initializer;
    _initializer = initializer;
    _reactive = reactive;
    if (changed && (active || reactive)) return reload();
    return SynchronousFuture(null);
  }

  Future<void> reload() {
    assert(_initializer != null, 'initializer not set');
    return load(_initializer!());
  }

  Future<void> load(Future<T> task) {
    setLoading();
    return task.then(setResult, onError: setError);
  }

  void setLoading() {
    _value = null;
    _state = ResultState.loading;
    invalidate();
  }

  void setResult(T value) {
    _value = value;
    _state = ResultState.success;
    invalidate();
  }

  void setError(Object error) {
    _value = error;
    _state = ResultState.error;
    invalidate();
  }

  void reset() {
    _value = null;
    _state = ResultState.idle;
    invalidate();
  }
}

enum ResultState {
  idle,
  loading,
  error,
  success,
}
