part of '../flutterx_live_data.dart';

mixin ObserverMixin {
  final Map<Object, Disposable> _disposables = {};
  Set<Object>? _keys;
  bool _dirty = false;

  @protected
  FutureOr<void> registerObservers();

  @protected
  void doRegister() {
    if (_keys != null) {
      _dirty = true;
      return;
    }
    final keys = _keys = {..._disposables.keys};
    final result = registerObservers();
    void doComplete() {
      for (final key in keys) _disposables.remove(key)!.dispose();
      _keys = null;
      if (_dirty) {
        _dirty = false;
        doRegister();
      }
    }

    result is Future ? result.whenComplete(doComplete) : doComplete();
  }

  @protected
  void doUnregister() {
    for (final key in [..._disposables.keys]) _disposables.remove(key)!.dispose();
  }

  void _observeAll(List<ObservableData> sources, Function srcOnChanged, void Function(List values) onChanged,
      Dispatcher? dispatcher) {
    final keys = _keys;
    if (keys == null) throw StateError('observe can only be called within registerObservers function');
    final key = _CombinedObserverKey(sources, srcOnChanged);
    _disposables.containsKey(key)
        ? keys.remove(key)
        : _disposables[key] = observeAllForever(sources, onChanged, dispatcher: dispatcher);
  }

  void _observe<V>(ObservableData<V> data, Observer<V> onChanged, Dispatcher? dispatcher) {
    final keys = _keys;
    if (keys == null) throw StateError('observe can only be called within registerObservers function');
    final key = _ObserverKey(data, onChanged);
    _disposables.containsKey(key)
        ? keys.remove(key)
        : _disposables[key] = data.observeForever(onChanged, dispatcher: dispatcher);
  }
}

@immutable
class _ObserverKey {
  final ObservableData _data;
  final Function _onChanged;

  const _ObserverKey(this._data, this._onChanged);

  @override
  bool operator ==(Object other) =>
      other is _ObserverKey && identical(other._data, _data) && other._onChanged == _onChanged;

  @override
  int get hashCode => Object.hashAll([_data, _onChanged]);
}

@immutable
class _CombinedObserverKey {
  final List<ObservableData> _data;
  final Function _onChanged;

  const _CombinedObserverKey(this._data, this._onChanged);

  @override
  bool operator ==(Object other) =>
      other is _CombinedObserverKey &&
      const ListEquality(IdentityEquality()).equals(other._data, _data) &&
      other._onChanged == _onChanged;

  @override
  int get hashCode => Object.hashAll([..._data, _onChanged]);
}
