part of '../flutterx_live_data.dart';

typedef StringEncode<T> = String Function(T value);
typedef StringDecode<T> = T Function(String source);

mixin StorageMixin<T> on ObservableData<T> {
  String get key;

  T get defaultValue;

  @protected
  Map<String, String> get storage;

  @protected
  StringEncode<T> get encode;

  @override
  void invalidate() {
    if (equality.equals(defaultValue, _data))
      storage.remove(key);
    else
      storage[key] = encode(_data);
    super.invalidate();
  }

  void reset() => _setData(defaultValue);

  static T resolve<T>(String key, T defaultValue, Map<String, String> storage, StringDecode<T> decode) {
    final value = storage[key];
    return value == null ? defaultValue : decode(value);
  }

  static String encodeEnum(dynamic value) => value == null ? '' : (value as Enum).name;

  static T decodeEnum<T>(String source, Iterable<T> values) =>
      (source.isEmpty ? null : (values as Iterable<Enum>).byName(source)) as T;

  static String encodeJsonIterable(Iterable value) => jsonEncode([...value]);

  static Iterable<T> decodeJsonIterable<T>(String source, T Function(dynamic item) decode) =>
      (jsonDecode(source) as Iterable).map(decode);

  static List<T> decodeList<T>(String source) => List<T>.from(jsonDecode(source));

  static Map<K, V> decodeMap<K, V>(String source) => Map<K, V>.from(jsonDecode(source));

  static String encodeEnumIterable(Iterable value) => encodeJsonIterable(value.map(encodeEnum));

  static Iterable<T> decodeEnumIterable<T>(String source, Iterable<T> values) =>
      decodeJsonIterable(source, (name) => decodeEnum<T>(name, values));
}
