part of '../flutterx_live_data.dart';

typedef LiveWidgetBuilder<T> = Widget Function(BuildContext context, T value);

class LiveDataBuilder<T> extends StatelessWidget {
  final ObservableData<T> data;
  final LiveWidgetBuilder<T> builder;
  final Widget? placeholder;

  const LiveDataBuilder({super.key, required this.data, required this.builder, this.placeholder});

  @override
  StatelessElement createElement() => _LiveDataBuilderElement<T>(this);

  @override
  Widget build(BuildContext context) =>
      data.initialized ? builder(context, data._data) : (placeholder ?? const SizedBox.shrink());

  @override
  void debugFillProperties(DiagnosticPropertiesBuilder properties) => super.debugFillProperties(properties
    ..add(DiagnosticsProperty<ObservableData<T>>('data', data))
    ..add(ObjectFlagProperty<LiveWidgetBuilder<T>>.has('builder', builder)));
}

class _LiveDataBuilderElement<T> extends StatelessElement {
  @override
  LiveDataBuilder<T> get widget => super.widget as LiveDataBuilder<T>;

  _LiveDataBuilderElement(LiveDataBuilder<T> super.widget);

  @override
  void mount(Element? parent, Object? newSlot) {
    widget.data.observeForever(_onChanged, dispatcher: Dispatcher.postFrame);
    super.mount(parent, newSlot);
  }

  @override
  void update(LiveDataBuilder<T> newWidget) {
    final oldWidget = widget;
    if (!identical(oldWidget.data, newWidget.data)) {
      oldWidget.data.removeObserver(_onChanged);
      newWidget.data.observeForever(_onChanged, dispatcher: Dispatcher.postFrame);
    }
    super.update(newWidget);
  }

  @override
  void unmount() {
    widget.data.removeObserver(_onChanged);
    super.unmount();
  }

  void _onChanged(_) => markNeedsBuild();
}
