part of '../flutterx_live_data.dart';

class CombinedLiveDataBuilder<T> extends StatelessWidget with MultipleLiveDataWidget {
  @override
  final List<ObservableData> data;
  final T Function(List values) transform;
  final LiveWidgetBuilder<T> builder;
  final Widget? placeholder;

  const CombinedLiveDataBuilder({
    super.key,
    required this.data,
    required this.transform,
    required this.builder,
    this.placeholder,
  });

  static CombinedLiveDataBuilder<Y> with2<X1, X2, Y>({
    Key? key,
    required ObservableData<X1> x1,
    required ObservableData<X2> x2,
    required Y Function(X1, X2) transform,
    required LiveWidgetBuilder<Y> builder,
    Widget? placeholder,
  }) =>
      CombinedLiveDataBuilder<Y>(
          data: [x1, x2],
          transform: (values) => transform(values[0], values[1]),
          builder: builder,
          placeholder: placeholder);

  static CombinedLiveDataBuilder<Y> with3<X1, X2, X3, Y>({
    Key? key,
    required ObservableData<X1> x1,
    required ObservableData<X2> x2,
    required ObservableData<X3> x3,
    required Y Function(X1, X2, X3) transform,
    required LiveWidgetBuilder<Y> builder,
    Widget? placeholder,
  }) =>
      CombinedLiveDataBuilder<Y>(
          data: [x1, x2, x3],
          transform: (values) => transform(values[0], values[1], values[2]),
          builder: builder,
          placeholder: placeholder);

  static CombinedLiveDataBuilder<Y> with4<X1, X2, X3, X4, Y>({
    Key? key,
    required ObservableData<X1> x1,
    required ObservableData<X2> x2,
    required ObservableData<X3> x3,
    required ObservableData<X4> x4,
    required Y Function(X1, X2, X3, X4) transform,
    required LiveWidgetBuilder<Y> builder,
    Widget? placeholder,
  }) =>
      CombinedLiveDataBuilder<Y>(
          data: [x1, x2, x3, x4],
          transform: (values) => transform(values[0], values[1], values[2], values[3]),
          builder: builder,
          placeholder: placeholder);

  static CombinedLiveDataBuilder<Y> with5<X1, X2, X3, X4, X5, Y>({
    Key? key,
    required ObservableData<X1> x1,
    required ObservableData<X2> x2,
    required ObservableData<X3> x3,
    required ObservableData<X4> x4,
    required ObservableData<X4> x5,
    required Y Function(X1, X2, X3, X4, X5) transform,
    required LiveWidgetBuilder<Y> builder,
    Widget? placeholder,
  }) =>
      CombinedLiveDataBuilder<Y>(
          data: [x1, x2, x3, x4, x5],
          transform: (values) => transform(values[0], values[1], values[2], values[3], values[4]),
          builder: builder,
          placeholder: placeholder);

  @override
  StatelessElement createElement() => _CombinedLiveDataBuilderElement<T>(this);

  @override
  Widget build(BuildContext context) {
    final data = context as _CombinedLiveDataBuilderElement<T>;
    return data._initialized ? builder(context, data._data) : (placeholder ?? const SizedBox.shrink());
  }

  @override
  void debugFillProperties(DiagnosticPropertiesBuilder properties) => super.debugFillProperties(properties
    ..add(IterableProperty<ObservableData>('data', data))
    ..add(ObjectFlagProperty<LiveWidgetBuilder<T>>.has('builder', builder))
    ..add(ObjectFlagProperty<T Function(List values)>.has('transform', transform)));
}

class _CombinedLiveDataBuilderElement<T> extends StatelessElement with MultipleLiveDataBuilderElementMixin {
  late T _data;
  bool _initialized = false;

  @override
  CombinedLiveDataBuilder<T> get widget => super.widget as CombinedLiveDataBuilder<T>;

  _CombinedLiveDataBuilderElement(CombinedLiveDataBuilder<T> super.widget);

  @override
  void updateData(Iterable<ObservableData> sources) {
    List values;
    try {
      values = ObservableData._allValues(sources);
    } catch (_) {
      _initialized = false; // at least one source is not initialized
      return;
    }
    _data = widget.transform(values);
    _initialized = true;
  }
}
