import 'package:flutter/material.dart';
import 'package:flutter_test/flutter_test.dart';
import 'package:flutterx_live_data/flutterx_live_data.dart';

void main() {
  group('Test LiveData', () {
    testWidgets('Test Storage', testStorageLiveData);
    testWidgets('Test History', testHistoryLiveData);
    testWidgets('Test Map', testLiveDataMap);
  });
}

Future<void> testStorageLiveData(WidgetTester tester) async {
  print(String.fromCharCode(0));
  final storage = <String, String>{};
  final ququq = StorageLiveQueue<int>(storage: storage, key: 'sss');
  ququq.add(17);
  print(ququq);
  ququq.remove(17);
  ququq.remove(333);
  print(ququq);
  final data = StorageMutableLiveData.string(key: 'my_data', storage: storage, defaultValue: 7);
  expect(data.value, 7);
  expect(storage['my_data'], null);
  data.value = 8;
  expect(storage['my_data'], '8');
}

Future<void> testHistoryLiveData(WidgetTester tester) async {
  final data = HistoryMutableLiveData(value: 175);
  expect(data.value, 175);
  data.value = 8;
  expect(data.value, 8);
  expect(data.canUndo.value, true);
  expect(data.canRedo.value, false);
  data.historyUndo();
  expect(data.value, 175);
  expect(data.canUndo.value, false);
  expect(data.canRedo.value, true);
  expect(data.index, 0);
  data
    ..value = 8
    ..value = 3
    ..value = 4
    ..value = 5
    ..index = 0;
  expect(data.value, 175);
  data.index = 1;
  expect(data.value, 8);
  data.index = 2;
  expect(data.value, 3);
  expect(data.canUndo.value, true);
  expect(data.canRedo.value, true);
  data
    ..historyReplace(5555)
    ..historyDelete()
    ..historyDelete()
    ..historyDelete();
  expect(data.value, 175);
  expect(data.canUndo.value, false);
  expect(data.canRedo.value, false);
  data.value = 13;
  expect(data.canUndo.value, false);
  data.historyReplace(988);
  expect(data.value, 988);
  expect(data.canUndo.value, false);
  expect(data.canRedo.value, false);
}

Future<void> testLiveDataMap(WidgetTester tester) async {
  final src = MutableLiveData(value: 0);
  final data = MediatorLiveData<String>()..addTransformSource(src, (value) => 'number $value');
  // Src value is initialized
  expect(src.value, 0);
  // Updating source value
  src.value++;
  await tester.pumpWidget(MaterialApp(builder: (context, _) {
    // Changing value during build
    src.value = 666;
    expect(src.value, 666);
    return LiveDataBuilder(data: data, builder: (context, value) => Text(value.toString()));
  }));
  // After build value has been dispatched
  expect(data.value, 'number 666');
  src.value = 777;
  // Value has been updated since computed value has active observers now
  expect(data.value, 'number 777');
  await tester.pumpFrames(
      MaterialApp(
          builder: (context, _) => CombinedLiveDataBuilder.with2<int, int, String>(
              x1: src,
              x2: src.transform((p0) => p0 * 2),
              transform: (a, b) => 'result: $a $b',
              builder: (context, value) => Text(value))),
      const Duration(seconds: 1));
  expect(find.text('result: 777 1554'), findsOneWidget);
}
